@@ -1,5 +1,6 @@
 /* locate -- search databases for filenames that match patterns
-   Copyright (C) 1994, 96, 98, 99, 2000, 2003 Free Software Foundation, Inc.
+   Copyright (C) 1994, 1996, 1998, 1999, 2000, 2003,
+                 2004, 2005 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +14,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
    USA.
 */
 
@@ -59,6 +60,12 @@
 #include <getopt.h>
 #include <xstrtol.h>
 
+#ifdef HAVE_UNISTD_H
+/* We need <unistd.h> for isatty(). */
+#include <unistd.h>
+#endif
+
+
 #define NDEBUG
 #include <assert.h>
 
@@ -113,6 +120,9 @@ extern int errno;
 #include "closeout.h"
 #include "nextelem.h"
 #include "regex.h"
+#include "quote.h"
+#include "quotearg.h"
+#include "printquoted.h"
 
 
 /* Note that this evaluates C many times.  */
@@ -135,22 +145,37 @@ static const char warn_name_units[] = N_("days");
 
 #define WARN_SECONDS ((SECONDS_PER_UNIT) * (WARN_NUMBER_UNITS))
 
+enum visit_result
+  {
+    VISIT_CONTINUE = 1,  /* please call the next visitor */
+    VISIT_ACCEPTED = 2,  /* accepted, call no futher callbacks for this file */
+    VISIT_REJECTED = 4,  /* rejected, process next file. */
+    VISIT_ABORT    = 8   /* rejected, process no more files. */
+  };
+
+enum ExistenceCheckType 
+  {
+    ACCEPT_EITHER,		/* Corresponds to lack of -E/-e option */
+    ACCEPT_EXISTING,		/* Corresponds to option -e */
+    ACCEPT_NON_EXISTING		/* Corresponds to option -E */
+  };
+
 /* Check for existence of files before printing them out? */
-static int check_existence = 0;
+enum ExistenceCheckType check_existence = ACCEPT_EITHER;
 
 static int follow_symlinks = 1;
 
 /* What to separate the results with. */
 static int separator = '\n';
 
-
-
+static struct quoting_options * quote_opts = NULL;
+static bool stdout_is_a_tty;
+static bool print_quoted_filename;
 
 /* Read in a 16-bit int, high byte first (network byte order).  */
 
 static short
-get_short (fp)
-     FILE *fp;
+get_short (FILE *fp)
 {
 
   register short x;
@@ -231,13 +256,12 @@ lc_strcpy(char *dest, const char *src)
   *dest = 0;
 }
 
-enum visit_result
-  {
-    VISIT_CONTINUE = 1,  /* please call the next visitor */
-    VISIT_ACCEPTED = 2,  /* accepted, call no futher callbacks for this file */
-    VISIT_REJECTED = 4,  /* rejected, process next file. */
-    VISIT_ABORT    = 8   /* rejected, process no more files. */
-  };
+struct locate_limits
+{
+  uintmax_t limit;
+  uintmax_t items_accepted;
+};
+static struct locate_limits limits;
 
 
 struct locate_stats
@@ -252,11 +276,20 @@ struct locate_stats
 static struct locate_stats statistics;
 
 
+struct stringbuf
+{
+  char *buffer;
+  size_t buffersize;
+  size_t *soffs;
+  size_t *preqlen;
+};
+static struct stringbuf casebuf;
+
+
 struct casefolder
 {
   const char *pattern;
-  char *buffer;
-  size_t buffersize;
+  struct stringbuf *pbuf;
 };
 
 struct regular_expression
@@ -280,25 +313,62 @@ struct visitor
 
 static struct visitor *inspectors = NULL;
 static struct visitor *lastinspector = NULL;
+static struct visitor *past_pat_inspector = NULL;
 
+/* 0 or 1 pattern(s) */
 static int
-process_filename(const char *munged_filename, const char *original_filename)
+process_simple(const char *munged_filename, const char *original_filename)
 {
   int result = VISIT_CONTINUE;
   const struct visitor *p = inspectors;
   
-  while ( (VISIT_CONTINUE == result) && (NULL != p) )
+  while ( ((VISIT_CONTINUE | VISIT_ACCEPTED) & result) && (NULL != p) )
     {
       result = (p->inspector)(munged_filename, original_filename, p->context);
       p = p->next;
     }
 
+    return result;
+}
+
+/* Accept if any pattern matches. */
+static int
+process_or (const char *munged_filename, const char *original_filename)
+{
+  int result = VISIT_CONTINUE;
+  const struct visitor *p = inspectors;
+  
+  while ( ((VISIT_CONTINUE | VISIT_REJECTED) & result) && (past_pat_inspector != p) )
+    {
+      result = (p->inspector)(munged_filename, original_filename, p->context);
+      p = p->next;
+    }
+
+  if (result == VISIT_CONTINUE)
+    result = VISIT_REJECTED;
+  if (result & (VISIT_ABORT | VISIT_REJECTED))
+    return result;
+
+  p = past_pat_inspector;
+  result = VISIT_CONTINUE;
+
+  while ( (VISIT_CONTINUE == result) && (NULL != p) )
+    {
+      result = (p->inspector)(munged_filename, original_filename, p->context);
+      p = p->next;
+    }
+  
   if (VISIT_CONTINUE == result)
     return VISIT_ACCEPTED;
   else
     return result;
 }
 
+typedef int (*processfunc)(const char *munged_filename,
+			    const char *original_filename);
+
+static processfunc mainprocessor = NULL;
+
 static void
 add_visitor(visitfunc fn, void *context)
 {
@@ -321,7 +391,19 @@ add_visitor(visitfunc fn, void *context)
 
 
 static int
-visit_justprint(const char *munged_filename, const char *original_filename, void *context)
+visit_justprint_quoted(const char *munged_filename, const char *original_filename, void *context)
+{
+  (void) context;
+  (void) munged_filename;
+  print_quoted (stdout, quote_opts, stdout_is_a_tty,
+		"%s", 
+		original_filename);
+  putchar(separator);
+  return VISIT_CONTINUE;
+}
+
+static int
+visit_justprint_unquoted(const char *munged_filename, const char *original_filename, void *context)
 {
   (void) context;
   (void) munged_filename;
@@ -331,8 +413,28 @@ visit_justprint(const char *munged_filename, const char *original_filename, void
 }
 
 static int
-visit_exists_follow(const char *munged_filename,
-		    const char *original_filename, void *context)
+visit_casefold(const char *munged_filename,
+	       const char *original_filename,
+	       void *context)
+{
+  struct stringbuf *b = context;
+  (void) original_filename;
+  
+  if (*b->preqlen+1 > b->buffersize)
+    {
+      b->buffer = xrealloc(b->buffer, *b->preqlen+1); /* XXX: consider using extendbuf(). */
+      b->buffersize = *b->preqlen+1;
+    }
+  lc_strcpy(b->buffer, munged_filename);
+
+  return VISIT_CONTINUE;
+}
+  
+/* visit_existing_follow implements -L -e */
+static int
+visit_existing_follow(const char *munged_filename,
+		      const char *original_filename,
+		      void *context)
 {
   struct stat st;
   (void) context;
@@ -353,9 +455,36 @@ visit_exists_follow(const char *munged_filename,
     }
 }
 
+/* visit_non_existing_follow implements -L -E */
 static int
-visit_exists_nofollow(const char *munged_filename,
-		      const char *original_filename, void *context)
+visit_non_existing_follow(const char *munged_filename,
+			  const char *original_filename,
+			  void *context)
+{
+  struct stat st;
+  (void) context;
+  (void) munged_filename;
+
+  /* munged_filename has been converted in some way (to lower case,
+   * or is just the base name of the file), and original_filename has not.  
+   * Hence only original_filename is still actually the name of the file 
+   * whose existence we would need to check.
+   */
+  if (stat(original_filename, &st) == 0)
+    {
+      return VISIT_REJECTED;
+    }
+  else
+    {
+      return VISIT_CONTINUE;
+    }
+}
+
+/* visit_existing_nofollow implements -P -e */
+static int
+visit_existing_nofollow(const char *munged_filename,
+			const char *original_filename,
+			void *context)
 {
   struct stat st;
   (void) context;
@@ -376,79 +505,109 @@ visit_exists_nofollow(const char *munged_filename,
     }
 }
 
+/* visit_non_existing_nofollow implements -P -E */
 static int
-visit_substring_match_nocasefold(const char *munged_filename, const char *original_filename, void *context)
+visit_non_existing_nofollow(const char *munged_filename,
+			    const char *original_filename,
+			    void *context)
+{
+  struct stat st;
+  (void) context;
+  (void) munged_filename;
+
+  /* munged_filename has been converted in some way (to lower case,
+   * or is just the base name of the file), and original_filename has not.  
+   * Hence only original_filename is still actually the name of the file 
+   * whose existence we would need to check.
+   */
+  if (lstat(original_filename, &st) == 0)
+    {
+      return VISIT_REJECTED;
+    }
+  else
+    {
+      return VISIT_CONTINUE;
+    }
+}
+
+static int
+visit_substring_match_nocasefold(const char *munged_filename,
+				 const char *original_filename,
+				 void *context)
 {
   const char *pattern = context;
   (void) original_filename;
 
   if (NULL != strstr(munged_filename, pattern))
-    return VISIT_CONTINUE;
+    return VISIT_ACCEPTED;
   else
     return VISIT_REJECTED;
 }
 
 static int
-visit_substring_match_casefold(const char *munged_filename, const char *original_filename, void *context)
+visit_substring_match_casefold(const char *munged_filename,
+			       const char *original_filename,
+			       void *context)
 {
-  struct casefolder * p = context;
-  size_t len = strlen(munged_filename);
-
+  const struct casefolder * p = context;
+  const struct stringbuf * b = p->pbuf;
+  (void) munged_filename;
   (void) original_filename;
-  if (len+1 > p->buffersize)
-    {
-      p->buffer = xrealloc(p->buffer, len+1); /* XXX: consider using extendbuf(). */
-      p->buffersize = len+1;
-    }
-  lc_strcpy(p->buffer, munged_filename);
-  
-  
-  if (NULL != strstr(p->buffer, p->pattern))
-    return VISIT_CONTINUE;
+
+  if (NULL != strstr(b->buffer, p->pattern))
+    return VISIT_ACCEPTED;
   else
     return VISIT_REJECTED;
 }
 
 
 static int
-visit_globmatch_nofold(const char *munged_filename, const char *original_filename, void *context)
+visit_globmatch_nofold(const char *munged_filename,
+		       const char *original_filename,
+		       void *context)
 {
   const char *glob = context;
   (void) original_filename;
   if (fnmatch(glob, munged_filename, 0) != 0)
     return VISIT_REJECTED;
   else
-    return VISIT_CONTINUE;
+    return VISIT_ACCEPTED;
 }
 
 
 static int
-visit_globmatch_casefold(const char *munged_filename, const char *original_filename, void *context)
+visit_globmatch_casefold(const char *munged_filename,
+			 const char *original_filename,
+			 void *context)
 {
   const char *glob = context;
   (void) original_filename;
   if (fnmatch(glob, munged_filename, FNM_CASEFOLD) != 0)
     return VISIT_REJECTED;
   else
-    return VISIT_CONTINUE;
+    return VISIT_ACCEPTED;
 }
 
 
 static int
-visit_regex(const char *munged_filename, const char *original_filename, void *context)
+visit_regex(const char *munged_filename,
+	    const char *original_filename,
+	    void *context)
 {
   struct regular_expression *p = context;
   (void) original_filename;
   
   if (0 == regexec(&p->re, munged_filename, 0u, NULL, 0))
-    return VISIT_CONTINUE;	/* match */
+    return VISIT_ACCEPTED;	/* match */
   else
     return VISIT_REJECTED;	/* no match */
 }
 
 
 static int
-visit_stats(const char *munged_filename, const char *original_filename, void *context)
+visit_stats(const char *munged_filename,
+	    const char *original_filename,
+	    void *context)
 {
   struct locate_stats *p = context;
   size_t len = strlen(original_filename);
@@ -488,7 +647,7 @@ visit_stats(const char *munged_filename, const char *original_filename, void *co
 /* Emit the statistics.
  */
 static void
-print_stats(size_t database_file_size)
+print_stats(int argc, size_t database_file_size)
 {
   char hbuf[LONGEST_HUMAN_READABLE + 1];
   
@@ -513,28 +672,33 @@ print_stats(size_t database_file_size)
 	 human_readable (statistics.highbit_filename_count,
 			 hbuf, human_ceiling, 1, 1));
   
-  printf(_("Compression ratio %4.2f%%\n"),
-	 100.0 * ((double)statistics.total_filename_length
-		  - (double) database_file_size)
-	 / (double) statistics.total_filename_length);
+  if (!argc)
+    printf(_("Compression ratio %4.2f%%\n"),
+	   100.0 * ((double)statistics.total_filename_length
+		    - (double) database_file_size)
+	   / (double) statistics.total_filename_length);
   printf("\n");
 }
 
 
-/* Print the entries in DBFILE that match shell globbing pattern PATHPART.
+/* Print the entries in DBFILE that match shell globbing patterns in ARGV.
    Return the number of entries printed.  */
 
 static unsigned long
-new_locate (char *pathpart,
-	    char *dbfile,
-	    int ignore_case,
-	    int enable_print,
-	    int basename_only,
-	    int use_limit,
-	    uintmax_t limit,
-	    int stats,
-	    int regex)
+locate (int argc,
+	char **argv,
+	char *dbfile,
+	int ignore_case,
+	int enable_print,
+	int basename_only,
+	int use_limit,
+	struct locate_limits *plimit,
+	int stats,
+	int regex)
 {
+  char *pathpart; 		/* A pattern to consider. */
+  int argn;			/* Index to current pattern in argv. */
+  int need_fold;	/* Set when folding and any pattern is non-glob. */
   FILE *fp;			/* The pathname database.  */
   int c;			/* An input byte.  */
   int nread;		     /* number of bytes read from an entry. */
@@ -544,14 +708,12 @@ new_locate (char *pathpart,
   int count = 0; /* The length of the prefix shared with the previous database entry.  */
   
   int old_format = 0; /* true if reading a bigram-encoded database.  */
+  struct visitor* pvis; /* temp for determining past_pat_inspector. */
   
   /* for the old database format,
      the first and second characters of the most common bigrams.  */
   char bigram1[128], bigram2[128];
 
-  /* number of items accepted (i.e. printed) */
-  unsigned long int items_accepted = 0uL;
-
   /* To check the age of the database.  */
   struct stat st;
   time_t now;
@@ -559,14 +721,32 @@ new_locate (char *pathpart,
   /* Set up the inspection regime */
   inspectors = NULL;
   lastinspector = NULL;
+  past_pat_inspector = NULL;
 
-  if (stats)
+
+  /* See if we need fold. */
+  if (ignore_case && !regex)
+    for ( argn = 0; argn < argc; argn++ )
+      {
+        pathpart = argv[argn];
+        if (!contains_metacharacter(pathpart))
+	  {
+	    need_fold = 1;
+	    break;
+	  }
+      }
+
+  if (need_fold)
     {
-      assert(!use_limit);
-      add_visitor(visit_stats, &statistics);
+      add_visitor(visit_casefold, &casebuf);
+      casebuf.preqlen = &pathsize;
+      casebuf.soffs = &count;
     }
-  else
+  
+  /* Add an inspector for each pattern we're looking for. */
+  for ( argn = 0; argn < argc; argn++ )
     {
+      pathpart = argv[argn];
       if (regex)
 	{
 	  struct regular_expression *p = xmalloc(sizeof(*p));
@@ -601,36 +781,66 @@ new_locate (char *pathpart,
 	    {
 	      struct casefolder * cf = xmalloc(sizeof(*cf));
 	      cf->pattern = pathpart;
-	      cf->buffer = NULL;
-	      cf->buffersize = 0;
+	      cf->pbuf = &casebuf;
 	      add_visitor(visit_substring_match_casefold, cf);
+	      /* If we ignore case, convert it to lower now so we don't have to
+	       * do it every time
+	       */
+	      lc_strcpy(pathpart, pathpart);
 	    }
 	  else
 	    {
 	      add_visitor(visit_substring_match_nocasefold, pathpart);
 	    }
 	}
+    }
 
-      /* We add visit_exists_*() as late as possible to reduce the
-       * number of stat() calls.
-       */
-      if (check_existence)
-	{
-	  visitfunc f;
-	  if (follow_symlinks)
-	    f = visit_exists_follow;
-	  else
-	    f = visit_exists_nofollow;
+  pvis = lastinspector;
+
+  /* We add visit_existing_*() as late as possible to reduce the
+   * number of stat() calls.
+   */
+  switch (check_existence)
+    {
+      case ACCEPT_EXISTING:
+	if (follow_symlinks)	/* -L, default */
+	  add_visitor(visit_existing_follow, NULL);
+	else			/* -P */
+	  add_visitor(visit_existing_nofollow, NULL);
+	break;
 	  
-	  add_visitor(f, NULL);
-	}
-      
+      case ACCEPT_NON_EXISTING:
+	if (follow_symlinks)	/* -L, default */
+	  add_visitor(visit_non_existing_follow, NULL);
+	else			/* -P */
+	  add_visitor(visit_non_existing_nofollow, NULL);
+	break;
 
-      if (enable_print)
-	add_visitor(visit_justprint, NULL);
+      case ACCEPT_EITHER:	/* Default, neither -E nor -e */
+	/* do nothing; no extra processing. */
+	break;
+    }
+      
+  if (stats)
+    add_visitor(visit_stats, &statistics);
+
+  if (enable_print)
+    {
+      if (print_quoted_filename)
+	add_visitor(visit_justprint_quoted,   NULL);
+      else
+	add_visitor(visit_justprint_unquoted, NULL);
     }
   
 
+  if (argc > 1)
+    {
+      past_pat_inspector = pvis->next;
+      mainprocessor = process_or;
+    }
+  else
+    mainprocessor = process_simple;
+
   if (stat (dbfile, &st) || (fp = fopen (dbfile, "r")) == NULL)
     {
       error (0, errno, "%s", dbfile);
@@ -670,18 +880,8 @@ new_locate (char *pathpart,
 	       old_format ? _("old") : "LOCATE02");
     }
   
-  /* If we ignore case, convert it to lower first so we don't have to
-   * do it every time
-   */
-  if (!stats && ignore_case)
-    {
-      lc_strcpy(pathpart, pathpart);
-    }
-  
-  items_accepted = 0;
-
   c = getc (fp);
-  while ( (c != EOF) && (!use_limit || (limit > 0)) )
+  while ( (c != EOF) && (!use_limit || (plimit->limit > 0)) )
     {
       register char *s;		/* Scan the path we read in.  */
 
@@ -736,9 +936,13 @@ new_locate (char *pathpart,
 	}
 
       testpath = basename_only ? base_name(path) : path;
-      if (VISIT_ACCEPTED == process_filename(testpath, path))
+
+      /* If we are searching for filename patterns, the inspector list 
+       * will contain an entry for each pattern for which we are searching.
+       */
+      if ((VISIT_ACCEPTED | VISIT_CONTINUE) & (mainprocessor)(testpath, path))
 	{
-	  if ((++items_accepted >= limit) && use_limit)
+	  if ((++plimit->items_accepted >= plimit->limit) && use_limit)
 	    {
 	      break;
 	    }
@@ -748,7 +952,7 @@ new_locate (char *pathpart,
       
   if (stats)
     {
-      print_stats(st.st_size);
+      print_stats(argc, st.st_size);
     }
   
   if (ferror (fp))
@@ -762,7 +966,7 @@ new_locate (char *pathpart,
       return 0;
     }
 
-  return items_accepted;
+  return plimit->items_accepted;
 }
 
 
@@ -774,15 +978,14 @@ extern char *version_string;
 char *program_name;
 
 static void
-usage (stream)
-     FILE *stream;
+usage (FILE *stream)
 {
   fprintf (stream, _("\
-Usage: %s [-d path | --database=path] [-e | --existing]\n\
+Usage: %s [-d path | --database=path] [-e | -E | --[non-]existing]\n\
       [-i | --ignore-case] [-w | --wholename] [-b | --basename] \n\
       [--limit=N | -l N] [-S | --statistics] [-0 | --null] [-c | --count]\n\
       [-P | -H | --nofollow] [-L | --follow] [-m | --mmap ] [ -s | --stdio ]\n\
-      [-r | --regex ] [--version] [--help] pattern...\n"),
+      [-p | --print] [-r | --regex ] [--version] [--help] pattern...\n"),
 	   program_name);
   fputs (_("\nReport bugs to <bug-findutils@gnu.org>.\n"), stream);
 }
@@ -791,6 +994,7 @@ static struct option const longopts[] =
 {
   {"database", required_argument, NULL, 'd'},
   {"existing", no_argument, NULL, 'e'},
+  {"non-existing", no_argument, NULL, 'E'},
   {"ignore-case", no_argument, NULL, 'i'},
   {"help", no_argument, NULL, 'h'},
   {"version", no_argument, NULL, 'v'},
@@ -799,6 +1003,7 @@ static struct option const longopts[] =
   {"wholename", no_argument, NULL, 'w'},
   {"wholepath", no_argument, NULL, 'w'}, /* Synonym. */
   {"basename", no_argument, NULL, 'b'},
+  {"print", no_argument, NULL, 'p'},
   {"stdio", no_argument, NULL, 's'},
   {"mmap",  no_argument, NULL, 'm'},
   {"limit",  required_argument, NULL, 'l'},
@@ -810,21 +1015,19 @@ static struct option const longopts[] =
 };
 
 int
-main (argc, argv)
-     int argc;
-     char **argv;
+main (int argc, char **argv)
 {
   char *dbpath;
   unsigned long int found = 0uL;
   int optc;
   int ignore_case = 0;
-  int print = 1;
+  int print = 0;
   int just_count = 0;
   int basename_only = 0;
-  uintmax_t limit = 0;
   int use_limit = 0;
   int regex = 0;
   int stats = 0;
+  char *e;
   
   program_name = argv[0];
 
@@ -835,17 +1038,24 @@ main (argc, argv)
   textdomain (PACKAGE);
   atexit (close_stdout);
 
+  limits.limit = 0;
+  limits.items_accepted = 0;
+
+  quote_opts = clone_quoting_options (NULL);
+  print_quoted_filename = true;
+  
   dbpath = getenv ("LOCATE_PATH");
   if (dbpath == NULL)
     dbpath = LOCATE_DB;
 
-  check_existence = 0;
+  check_existence = ACCEPT_EITHER;
 
-  while ((optc = getopt_long (argc, argv, "bcd:eil:rsm0SwHPL", longopts, (int *) 0)) != -1)
+  while ((optc = getopt_long (argc, argv, "bcd:eEil:prsm0SwHPL", longopts, (int *) 0)) != -1)
     switch (optc)
       {
       case '0':
 	separator = 0;
+	print_quoted_filename = false; /* print filename 'raw'. */
 	break;
 
       case 'b':
@@ -854,7 +1064,6 @@ main (argc, argv)
 
       case 'c':
 	just_count = 1;
-	print = 0;
 	break;
 
       case 'd':
@@ -862,7 +1071,11 @@ main (argc, argv)
 	break;
 
       case 'e':
-	check_existence = 1;
+	check_existence = ACCEPT_EXISTING;
+	break;
+
+      case 'E':
+	check_existence = ACCEPT_NON_EXISTING;
 	break;
 
       case 'i':
@@ -873,6 +1086,10 @@ main (argc, argv)
 	usage (stdout);
 	return 0;
 
+      case 'p':
+	print = 1;
+	break;
+
       case 'v':
 	printf (_("GNU locate version %s\n"), version_string);
 	return 0;
@@ -906,7 +1123,7 @@ main (argc, argv)
       case 'l':
 	{
 	  char *end = optarg;
-	  strtol_error err = xstrtoumax(optarg, &end, 10, &limit, NULL);
+	  strtol_error err = xstrtoumax(optarg, &end, 10, &limits.limit, NULL);
 	  if (LONGINT_OK != err)
 	    {
 	      STRTOL_FATAL_ERROR(optarg, _("argument to --limit"), err);
@@ -927,56 +1144,57 @@ main (argc, argv)
 	return 1;
       }
 
+  if (!just_count && !stats)
+    print = 1;
+
   if (stats)
     {
-      use_limit = 0;
-      print = 0;
+      if (optind == argc)
+ 	  use_limit = 0;
     }
   else
     {
-      if (optind == argc)
+      if (!just_count && optind == argc)
 	{
 	  usage (stderr);
 	  return 1;
 	}
     }
   
-  for (; stats || optind < argc; optind++)
+
+  if (1 == isatty(STDOUT_FILENO))
+    stdout_is_a_tty = true;
+  else
+    stdout_is_a_tty = false;
+
+  next_element (dbpath, 0);	/* Initialize.  */
+  while ((e = next_element ((char *) NULL, 0)) != NULL)
     {
-      char *e;
-      const char *needle;
-      next_element (dbpath, 0);	/* Initialize.  */
-      needle = stats ? NULL : argv[optind];
-      while ((e = next_element ((char *) NULL, 0)) != NULL)
+      statistics.compressed_bytes = 
+      statistics.total_filename_count = 
+      statistics.total_filename_length = 
+      statistics.whitespace_count = 
+      statistics.newline_count = 
+      statistics.highbit_filename_count = 0u;
+
+      if (0 == strlen(e) || 0 == strcmp(e, "."))
 	{
-	  statistics.compressed_bytes = 
-	    statistics.total_filename_count = 
-	    statistics.total_filename_length = 
-	    statistics.whitespace_count = 
-	    statistics.newline_count = 
-	    statistics.highbit_filename_count = 0u;
-
-	  if (0 == strlen(e) || 0 == strcmp(e, "."))
-	    {
-	      /* Use the default database name instead (note: we
-	       * don't use 'dbpath' since that might itself contain a 
-	       * colon-separated list.
-	       */
-	      e = LOCATE_DB;
-	    }
-	  
-	  found += new_locate (needle, e, ignore_case, print, basename_only, use_limit, limit, stats, regex);
+	  /* Use the default database name instead (note: we
+	   * don't use 'dbpath' since that might itself contain a 
+	   * colon-separated list.
+	   */
+	  e = LOCATE_DB;
 	}
-      if (stats)
-	break;
+	  
+      found = locate (argc - optind, &argv[optind], e, ignore_case, print, basename_only, use_limit, &limits, stats, regex);
     }
-
+  
   if (just_count)
     {
       printf("%ld\n", found);
     }
   
-  if (found || (use_limit && (limit==0)) || stats )
+  if (found || (use_limit && (limits.limit==0)) || stats )
     return 0;
   else
     return 1;
